from enum import IntEnum
from typing import Any, Dict, List, Literal, Union

import vapoursynth as vs

core = vs.core


__all__ = ['SAMPLEMODE', 'F3kdb', 'SampleMode']


SAMPLEMODE = Literal[1, 2, 3, 4]


class SampleMode(IntEnum):
    COLUMN = 1
    SQUARE = 2
    ROW = 3
    COL_ROW_MEAN = 4


class F3kdb:
    """f3kdb object."""
    radius: int
    thy: int
    thcb: int
    thcr: int
    gry: int
    grc: int
    sample_mode: int
    use_neo: bool
    f3kdb_args: Dict[str, Any]

    _step: int

    def __init__(self,
                 radius: int = 16,
                 threshold: Union[int, List[int]] = 30, grain: Union[int, List[int]] = 0,
                 sample_mode: Union[SAMPLEMODE, SampleMode] = 2, use_neo: bool = False, **kwargs: Any) -> None:
        """
        Handle debanding operations onto a clip using a set of configured parameters.

        Both f3kdb and neo_f3kdb actually change their strength at 1 + 16 * n for SampleMode.SQUARE
        and 1 + 32 * n for SampleMode.COLUMN, SampleMode.ROW or SampleMode.COL_ROW_MEAN.
        This function is aiming to average n and n + 1 strength for a better accuracy.
        Original function written by Z4ST1N, modified by Vardë.

        :param radius:          Banding detection range
        :param threshold:       Banding detection threshold(s) for planes
        :param grain:           Specifies amount of grains added in the last debanding stage
        :param sample_mode:     Valid modes are:
                                * SampleMode.COLUMN or 1: Take 2 pixels as reference pixel
                                  Reference pixels are in the same column of current pixel
                                * SampleMode.SQUARE or 2: Take 4 pixels as reference pixel
                                  Reference pixels are in the square around current pixel
                                * SampleMode.ROW or 3: Take 2 pixels as reference pixel
                                  Reference pixels are in the same row of current pixel
                                  Only `neo_f3kdb.Deband` supports it
                                * SampleMode.COL_ROW_MEAN or 4: Arithmetic mean of 1 and 3
                                  Reference points are randomly picked within the range
                                  Only `neo_f3kdb.Deband` supports it
        :param use_neo:         Use `neo_f3kdb.Deband`
        :param kwargs:          Arguments passed to f3kdb.Deband
                                Default are `keep_tv_range=True, output_depth=16`
                                Read the f3kdb's documentation for more information about them:
                                https://f3kdb.readthedocs.io/en/latest/usage.html#parameters
        """
        self.radius = radius

        th_s = [threshold] * 3 if isinstance(threshold, int) else threshold + [threshold[-1]] * (3 - len(threshold))
        self.thy, self.thcb, self.thcr = [max(1, x) for x in th_s]

        self.gry, self.grc = [grain] * 2 if isinstance(grain, int) else grain + [grain[-1]] * (2 - len(grain))

        if sample_mode > 2 and not use_neo:
            raise ValueError('F3kdb: f3kdb.Deband doesn\'t support SampleMode.ROW or SampleMode.COL_ROW_MEAN')

        self.sample_mode = sample_mode
        self.use_neo = use_neo

        self._step = 16 if sample_mode == 2 else 32

        self.f3kdb_args = dict(keep_tv_range=True, output_depth=16)
        self.f3kdb_args |= kwargs

    def deband(self, clip: vs.VideoNode) -> vs.VideoNode:
        """
        Main deband function

        :param clip:            Source clip
        :return:                Debanded clip
        """
        if clip.format is None:
            raise ValueError('deband: Variable format not allowed!')

        if self.thy % self._step == 1 and self.thcb % self._step == 1 and self.thcr % self._step == 1:
            deband = self._pick_f3kdb(self.use_neo,
                                      clip, self.radius,
                                      self.thy, self.thcb, self.thcr,
                                      self.gry, self.grc,
                                      self.sample_mode, **self.f3kdb_args)
        else:
            loy, locb, locr = [(th - 1) // self._step * self._step + 1 for th in [self.thy, self.thcb, self.thcr]]
            hiy, hicb, hicr = [lo + self._step for lo in [loy, locb, locr]]

            lo_clip = self._pick_f3kdb(self.use_neo,
                                       clip, self.radius,
                                       loy, locb, locr,
                                       self.gry, self.grc,
                                       self.sample_mode, **self.f3kdb_args)
            hi_clip = self._pick_f3kdb(self.use_neo,
                                       clip, self.radius,
                                       hiy, hicb, hicr,
                                       self.gry, self.grc,
                                       self.sample_mode, **self.f3kdb_args)

            if clip.format.color_family == vs.GRAY:
                weight = [
                    (self.thy - loy) / self._step
                ]
            else:
                weight = [
                    (self.thy - loy) / self._step,
                    (self.thcb - locb) / self._step,
                    (self.thcr - locr) / self._step
                ]

            deband = core.std.Merge(lo_clip, hi_clip, weight)

        return deband

    def grain(self, clip: vs.VideoNode) -> vs.VideoNode:
        """
        Convenience function that set thresholds to 1 (basically it doesn't deband)

        :param clip:            Source clip
        :return:                Grained clip
        """
        self.thy, self.thcr, self.thcb = (1, ) * 3
        return self.deband(clip)

    @staticmethod
    def _pick_f3kdb(neo: bool, *args: Any, **kwargs: Any) -> vs.VideoNode:
        return core.neo_f3kdb.Deband(*args, **kwargs) if neo else core.f3kdb.Deband(*args, **kwargs)



from typing import Any, Dict, List, Union

import vapoursynth as vs
from vsutil import join, split

core = vs.core


__all__ = ['Placebo']


class Placebo:
    radius: float
    thrs: List[float]
    iterations: int
    grain: List[float]

    placebodb_args: Dict[str, Any]

    def __init__(self,
                 radius: float = 16.0, threshold: Union[float, List[float]] = 4.0,
                 iterations: int = 1, grain: Union[float, List[float]] = 6.0,
                 **kwargs: Any) -> None:
        """
        Wrapper for placebo.Deband
        https://github.com/Lypheo/vs-placebo#vs-placebo

        :param radius:          The debanding filter's initial radius.
                                The radius increases linearly for each iteration.
                                A higher radius will find more gradients,
                                but a lower radius will smooth more aggressively.

        :param threshold:       The debanding filter's cut-off threshold.
                                Higher numbers increase the debanding strength dramatically,
                                but progressively diminish image details.

        :param iterations:      The number of debanding steps to perform per sample.
                                Each step reduces a bit more banding,
                                but takes time to compute.
                                Note that the strength of each step falls off very quickly,
                                so high numbers (>4) are practically useless.

        :param grain:           Add some extra noise to the image.
                                This significantly helps cover up remaining quantization artifacts.
                                Higher numbers add more noise.
                                Note: When debanding HDR sources, even a small amount of grain can result
                                in a very big change to the brightness level.
                                It's recommended to either scale this value down or disable it entirely for HDR.

        :param kwargs:          Arguments passed to `placebo.Deband`.
        """

        self.radius = radius

        if isinstance(threshold, (float, int)):
            self.threshold = [threshold] * 3
        else:
            self.threshold = threshold + [threshold[-1]] * (3 - len(threshold))

        self.iterations = iterations

        if isinstance(grain, (float, int)):
            self.grain = [grain] * 3
        else:
            self.grain = grain + [grain[-1]] * (3 - len(grain))

        self.placebodb_args = kwargs

    def deband(self, clip: vs.VideoNode) -> vs.VideoNode:
        """
        Main deband function

        :param clip:            Source clip
        :return:                Debanded clip
        """
        if clip.format is None:
            raise ValueError('deband: Variable format not allowed!')

        debs = [
            p.placebo.Deband(1, self.iterations, thr, self.radius, gra, **self.placebodb_args)
            for p, thr, gra in zip(split(clip), self.threshold, self.grain)
        ]
        return debs[0] if len(debs) == 1 else join(debs, clip.format.color_family)



# flake8: noqa

from .debanders import *
from .f3kdb import *
from .placebo import *

__all__ = [
    'dumb3kdb', 'f3kbilateral', 'f3kpf', 'lfdeband', 'placebo_deband',
    'SAMPLEMODE', 'F3kdb', 'SampleMode',
    'Placebo'
]



"""
    Various functions used for debanding.

    This used to be the `debandshit` module written by Z4ST1N,
    with some functions that were rarely (if ever) used removed because I can't reasonably maintain them.
"""
from typing import Any, Dict, List, Optional, Union

import vapoursynth as vs
from vsutil import depth

from .f3kdb import SAMPLEMODE, F3kdb, SampleMode
from .placebo import Placebo

core = vs.core

__all__ = ['dumb3kdb', 'f3kbilateral', 'f3kpf', 'lfdeband', 'placebo_deband']


def dumb3kdb(clip: vs.VideoNode, radius: int = 16,
             threshold: Union[int, List[int]] = 30, grain: Union[int, List[int]] = 0,
             sample_mode: Union[SAMPLEMODE, SampleMode] = 2, use_neo: bool = False, **kwargs: Any) -> vs.VideoNode:
    """Small convenience function for calling F3kdb().deband()."""
    return F3kdb(radius, threshold, grain, sample_mode, use_neo, **kwargs).deband(clip)


def f3kbilateral(clip: vs.VideoNode, radius: int = 16,
                 threshold: Union[int, List[int]] = 65, grain: Union[int, List[int]] = 0,
                 f3kdb_args: Optional[Dict[str, Any]] = None,
                 limflt_args: Optional[Dict[str, Any]] = None) -> vs.VideoNode:
    """
    f3kbilateral: f3kdb multistage bilateral-esque filter from debandshit.

    This function is more of a last resort for extreme banding.
    Recommend values are ~40-60 for y and c strengths.

    Dependencies:
    * mvsfunc

    :param clip:        Input clip
    :param radius:      Same as F3kdb constructor.
    :param threshold:   Same as F3kdb constructor.
    :param grain:       Same as F3kdb constructor.
                        It happens after mvsfunc.LimitFilter
                        and call another instance of F3kdb if != 0.
    :f3kdb_args:        Same as F3kdb kwargs constructor.
    :lf_args:           Arguments passed to mvsfunc.LimitFilter.

    :return:            Debanded clip
    """
    try:
        from mvsfunc import LimitFilter
    except ModuleNotFoundError as mod_err:
        raise ModuleNotFoundError("f3kbilateral: missing dependency 'mvsfunc'") from mod_err

    if clip.format is None:
        raise ValueError("f3kbilateral: 'Variable-format clips not supported'")

    bits = clip.format.bits_per_sample

    f3_args: Dict[str, Any] = dict()
    if f3kdb_args is not None:
        f3_args |= f3kdb_args

    lf_args: Dict[str, Any] = dict(thr=0.6, elast=3.0, thrc=None)
    if limflt_args is not None:
        lf_args |= limflt_args

    rad1 = round(radius * 4 / 3)
    rad2 = round(radius * 2 / 3)
    rad3 = round(radius / 3)

    db1 = F3kdb(rad1, threshold, 0, **f3_args)
    db2 = F3kdb(rad2, threshold, 0, **f3_args)
    db3 = F3kdb(rad3, threshold, 0, **f3_args)

    # Edit the thr of first f3kdb object
    db1.thy, db1.thcb, db1.thcr = [max(1, th // 2) for th in (db1.thy, db1.thcb, db1.thcr)]

    clip = depth(clip, 16)

    flt1 = db1.deband(clip)
    flt2 = db2.deband(flt1)
    flt3 = db3.deband(flt2)

    limit = LimitFilter(flt3, flt2, ref=clip, **lf_args)

    if grain:
        grained = F3kdb(grain=grain, **f3_args).grain(limit)
    else:
        grained = limit

    return depth(grained, bits)


def f3kpf(clip: vs.VideoNode, radius: int = 16,
          threshold: Union[int, List[int]] = 30, grain: Union[int, List[int]] = 0,
          f3kdb_args: Optional[Dict[str, Any]] = None,
          limflt_args: Optional[Dict[str, Any]] = None) -> vs.VideoNode:
    """
    f3kdb with a simple prefilter by mawen1250 - https://www.nmm-hd.org/newbbs/viewtopic.php?f=7&t=1495#p12163.

    Since the prefilter is a straight gaussian+average blur, f3kdb's effect becomes very strong, very fast.
    Functions more or less like gradfun3 without the detail mask.

    Dependencies:
    * mvsfunc

    :param clip:        Input clip
    :param radius:      Banding detection range
    :param threshold:   Banding detection thresholds for multiple planes
    :param f3kdb_args:  Arguments passed to F3kdb constructor
    :param limflt_args: Arguments passed to mvsfunc.LimitFilter

    :return:            Debanded clip
    """
    try:
        from mvsfunc import LimitFilter
    except ModuleNotFoundError as mod_err:
        raise ModuleNotFoundError("f3kpf: missing dependency 'mvsfunc'") from mod_err

    if clip.format is None:
        raise ValueError("f3kpf: 'Variable-format clips not supported'")

    f3_args: Dict[str, Any] = dict()
    if f3kdb_args is not None:
        f3_args |= f3kdb_args

    lf_args: Dict[str, Any] = dict(thr=0.3, elast=2.5, thrc=None)
    if limflt_args is not None:
        lf_args |= limflt_args

    blur = core.std.Convolution(clip, [1, 2, 1, 2, 4, 2, 1, 2, 1]).std.Convolution([1] * 9, planes=0)
    diff = core.std.MakeDiff(clip, blur)

    deband = F3kdb(radius, threshold, grain, **f3_args).deband(blur)
    deband = LimitFilter(deband, blur, **lf_args)

    return core.std.MergeDiff(deband, diff)


def lfdeband(clip: vs.VideoNode) -> vs.VideoNode:
    """
    A simple debander ported from AviSynth.

    :param clip:        Input clip

    :return:            Debanded clip
    """
    if clip.format is None:
        raise ValueError("lfdeband: 'Variable-format clips not supported'")

    bits = clip.format.bits_per_sample
    wss, hss = 1 << clip.format.subsampling_w, 1 << clip.format.subsampling_h
    w, h = clip.width, clip.height
    dw, dh = round(w / 2), round(h / 2)

    clip = depth(clip, 16)
    dsc = core.resize.Spline64(clip, dw-dw % wss, dh-dh % hss)

    d3kdb = F3kdb(radius=30, threshold=80, grain=0).deband(clip)

    ddif = core.std.MakeDiff(d3kdb, dsc)

    dif = core.resize.Spline64(ddif, w, h)
    out = core.std.MergeDiff(clip, dif)
    return depth(out, bits)


def placebo_deband(clip: vs.VideoNode, radius: float = 16.0, threshold: Union[float, List[float]] = 4.0,
                   iterations: int = 1, grain: Union[float, List[float]] = 6.0, **kwargs: Any) -> vs.VideoNode:
    """Small convenience function for calling Placebo().deband()."""
    return Placebo(radius, threshold, iterations, grain, **kwargs).deband(clip)



from setuptools import setup

with open("README.md") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    install_requires = fh.read()

NAME = "vs-debandshit"

RELEASE = "0.4.5"

setup(
    name=NAME,
    version=RELEASE,
    author="LightArrowsEXE",
    author_email="Lightarrowsreboot@gmail.com",
    description="VapourSynth Debanding Functions Collection",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["debandshit"],
    url="https://github.com/LightArrowsEXE/debandshit",
    package_data={
        'debandshit': ['py.typed'],
    },
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    python_requires='>=3.9',
)
